function err = arabic_speech_experiment(representation, classifier, states)
%ARABIC_SPEECH_EXPERIMENT Classification experiment on Arabic speech data
%
%   err = arabic_speech_experiment(type, states)
%
% The function performs a time series embedding experiment on the Arabic
% speech digits data set. The variable type can have values 'hmm' for
% an HMM Bayes classifier, 'hmm_softmax' for a softmax trained on HMM
% scores, 'fisher' for Fisher kernel SVMs, or 'fkl' for supervised time 
% series embedding (default = 'fkl'). You can also specify the number of 
% HMM states for which to run the experiments.
% The function returns a bunch of generalization errors in a cell array
% with one cell per number of states.
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~exist('states', 'var') || isempty(states)
        states = 10;
    end
    if ~exist('classifier', 'var') || isempty(classifier)
        classifier = 'all';
    end    
    if ischar(classifier) 
        if strcmpi(classifier, 'all')
            classifier = {'bayes', 'softmax', 'svm', 'lmnn'};
        else
            classifier = {classifier};
        end
    end 
    if strcmpi(representation, 'dtw') && ~strcmpi(classifier, '1nn')
        warning('DTW distances can only be classified using 1-NN.');
        classifier = '1nn';
    end    
    
    % Load data
    addpath('libsvm-2.9');
    rmpath([matlabroot '/toolbox/stats']);
    addpath('prtools'); 
    addpath(genpath('drtoolbox'));
    train_X = []; test_X = [];
    load 'data/arabic_speech.mat'
    train_labels = train_T;
    test_labels  = test_T;
    
    % Initialize some variables
    err = ones(length(states), 1);
    
    % Perform experiments for different numbers of HMM states
    for index=1:length(states)
        
        % Use fixed training/test division
        no_states = states(index);
        N = length(train_X);
        O = length(test_X); 
        D = size(train_X{1}, 1); 
        K = no_states;

        % Construct time series representation
        switch representation

            case 'hmm_posterior'

                % Train HMM for each class 
                no_labels = length(unique(train_labels));
                model = cell(no_labels, 1);
                for i=1:no_labels
                    disp(['Training HMM ' num2str(i) ' of ' num2str(no_labels) '...']);
                    model{i} = train_hmm(train_X(train_labels == i), no_states, 'gaussian');
                end

                % Perform Viterbi decoding on each sequence using each HMM
                train_feat = zeros(length(train_X), no_labels);                    
                test_feat  = zeros(length(test_X),  no_labels);
                disp('Running Viterbi decoders...');
                for i=1:length(train_X)
                    for j=1:length(model)
                        [foo, train_feat(i, j)] = viterbi(train_X{i}, model{j});
                    end
                end                    
                for i=1:length(test_X)
                    for j=1:length(model)
                        [foo, test_feat(i, j)] = viterbi(test_X{i}, model{j});
                    end
                end

            case 'hmm_gradient'

                % Train single HMM on data set
                include_pi = true; include_A = true; include_mu = true; include_sigma = false; include_mix = false;                    
                model = train_hmm(train_X, no_states, 'gaussian');

                % Compute gradient representations
                train_feat = zeros(length(train_X), include_pi * K + include_A * K * K + include_mu * D * K);
                test_feat  = zeros(length(test_X),  include_pi * K + include_A * K * K + include_mu * D * K);
                for i=1:length(train_X)
                    [foo, tmp] = hmm_grad([model.pi(:); model.A(:); model.mu(:); model.sigma(:)], train_X{i}, model, include_pi, include_A, include_mu, include_sigma, include_mix);
                    train_feat(i,:) = tmp';
                end        
                for i=1:length(test_X)
                    [foo, tmp] = hmm_grad([model.pi(:); model.A(:); model.mu(:); model.sigma(:)], test_X{i}, model, include_pi, include_A, include_mu, include_sigma, include_mix);
                    test_feat(i,:) = tmp;
                end                        

            case 'fkl'

                % Cross-validate over lambda in first iterations???
                include_pi = true; include_A = true; include_mu = true; include_sigma = false; include_mix = false;                                        
                addpath('minFunc');
                lambda = 0;

                % Train HMM using STSE
                [train_feat, model, M, mean_X, var_X] = stse_metric(train_X, train_labels, no_states, lambda, include_pi, include_A, include_mu);
                train_feat = train_feat'; 
                M = M';

                % Compute features for test data
                test_feat = zeros(length(test_X), include_pi * K + include_A * K * K + include_mu * D * K);
                for i=1:O
                    Xi = bsxfun(@rdivide, bsxfun(@minus, test_X{i}, mean_X), sqrt(var_X));
                    [~, test_feat(i,:)] = hmm_grad([model.pi(:); model.A(:); model.mu(:); model.sigma(:)], Xi, model, include_pi, include_A, include_mu, include_sigma, include_mix);
                end

                % Apply metric on both representations
                train_feat = bsxfun(@times, train_feat, sqrt(M));
                test_feat  = bsxfun(@times,  test_feat, sqrt(M));
                
            case 'dtw'
                
                % Compute dynamic time warping distances
                D = zeros(length(train_X), length(test_X));
                for i=1:length(train_X)
                    if ~rem(i, 250)
                        disp(['Matrix row ' num2str(i) ' of ' num2str(length(train_X))]);
                    end
                    for j=1:length(test_X)
                        [D(i, j), ~, norm] = dtw(train_X{i}, test_X{j});
                        D(i, j) = D(i, j) ./ norm;
                    end
                end

            otherwise
                error('Unknown representation.');
        end

        % Perform classification
        if ischar(classifier)
            classifier = {classifier};
        end
        for i=1:length(classifier)
            
            % Select current classifier
            classif = classifier{i};
            switch classif
                case 'bayes'

                    if ~strcmpi(representation, 'hmm_posterior')
                        warning('The Bayes classifier can only be used in combination with the HMM posterior representation. Returning error of 1.');
                        est_labels = test_labels + 1;
                    else

                        % Multiply with class prior
                        prior = zeros(1, no_labels);
                        for j=1:no_labels
                            prior(j) = sum(train_labels == j);
                        end
                        prior = prior ./ sum(prior);
                        test_feat = bsxfun(@plus, test_feat, log(prior));

                        % Perform MAP classification
                        [foo, est_labels] = max(test_feat, [], 2);
                    end

                case 'softmax'

                    % Cross-validation to determine value of L2-regularizer
                    test_feat  = test_feat  ./ (max(train_feat(:)));
                    train_feat = train_feat ./ (max(train_feat(:)));
                    lambda = [0 0.0001 0.0002 0.0005 0.001 0.002 0.005 0.01 0.02 0.05 0.1 0.2 0.5 1 2 5];
                    lambda_err = nan(length(lambda), 1);
                    valid_ind = randperm(N)';
                    trn_valid_ind = valid_ind(1:round(.9 * N));
                    tst_valid_ind = valid_ind(1+round(.9 * N):end);                    
                    for j=1:length(lambda)
                        disp(['Training logistic regressor ' num2str(j) ' of ' num2str(length(lambda)) '...']);
                        [E, E_bias] = train_logreg(train_feat(trn_valid_ind,:)', train_labels(trn_valid_ind), lambda(j));
                        est_labels = eval_logreg(train_feat(tst_valid_ind,:)', E, E_bias)';
                        lambda_err(j) = sum(est_labels ~= train_labels(tst_valid_ind)) ./ numel(tst_valid_ind);
                    end
                    [foo, min_ind] = min(lambda_err);
                    lambda = lambda(min_ind);
                    disp(['Using value of lambda of ' num2str(lambda) '...']);                    

                    % Train and test a softmax on the posterior probabilities
                    [E, E_bias] = train_logreg(train_feat', train_labels, lambda);
                    est_labels = eval_logreg(test_feat', E, E_bias)';

                case 'svm'

                    % Cross-validation to determine value of slack variable
                    disp('Performing cross-validation over slack variable...');
                    C = 0.25:.25:2.5;
                    C_err = nan(length(C), 1);
                    valid_ind = randperm(N)';
                    trn_valid_ind = valid_ind(1:round(.9 * N));
                    tst_valid_ind = valid_ind(1+round(.9 * N):end);                    
                    trn_valid_feat = train_feat(trn_valid_ind,:);
                    tst_valid_feat = train_feat(tst_valid_ind,:);
                    train_K = dataset(trn_valid_feat * trn_valid_feat', train_labels(trn_valid_ind));
                    test_K  = dataset(tst_valid_feat * trn_valid_feat', train_labels(tst_valid_ind));
                    for j=1:length(C)
                        disp(['Training SVM ' num2str(j) ' of ' num2str(length(C)) '...']);
                        svm = libsvc(train_K, 0, C(j));
                        est_labels = labeld(map(test_K, svm));
                        C_err(j) = sum(est_labels ~= train_labels(tst_valid_ind)) ./ numel(tst_valid_ind);
                    end
                    [foo, min_ind] = min(C_err);
                    C = C(min_ind);
                    disp(['Using value of C of ' num2str(C) '...']);

                    % Compute (practical) Fisher kernel
                    disp(['Performing final SVM training (slack = ' num2str(C) ')...']);
                    train_K = dataset(train_feat * train_feat', train_labels);
                    test_K  = dataset(test_feat  * train_feat',  test_labels);

                    % Train and test final Fisher kernel SVM
                    svm = libsvc(train_K, 0, C);
                    est_labels = labeld(map(test_K, svm));

                case 'lmnn'

                    % Perform PCA + LMNN
                    addpath(genpath('mLMNN'));
                    [train_feat2, mapping] = compute_mapping(train_feat, 'PCA', min(min(size(train_feat)), 200)); 
                    metric = lmnn2(train_feat2', train_labels', 5, 'quiet', 1, 'maxiter', 500, 'validation', 0.1);

                    % Perform classification of test data 
                    test_feat2 = out_of_sample(test_feat, mapping);
                    [foo, est_labels] = energyclassify(metric, train_feat2', train_labels', test_feat2', test_labels', 9); % should cross-validate over k?
                    est_labels = est_labels';
                    
                case '1nn'          % assumes we have distance matrix D

                    % Perform 1-NN classifications
                    [~, ii] = sort(D, 1, 'ascend');
                    est_labels = train_labels(ii(1,:));

                otherwise
                    error('Unknown classifier.');                    
            end

            % Measure generalization error on test set
            err(index) = sum(est_labels ~= test_labels) ./ numel(test_labels);
            disp(['Generalization error on test set (' classif ' classifier): ' num2str(err(index))]); 
        end
    end        
    