function [mappedX, network, binary_data, mean_X, var_X] = train_autoencoder(X, layers)
%TRAIN_AUTOENCODER Trains an autoencoder using RBM pretraining
%
%   [mappedX, network] = train_encoder(X, layers)
%
% Trains up an autoencoder with the structure that is specified in layers. 
% The low-dimensional data is returned in mappedX, and the network in
% network. This function is mainly designed to be used for binary [0,1]
% data.

% This file is part of the Matlab Toolbox for Dimensionality Reduction v0.7.2b.
% The toolbox can be obtained from http://homepage.tudelft.nl/19j49
% You are free to use, change, or redistribute this code in any way you
% want for non-commercial purposes. However, it is appreciated if you 
% maintain the name of the original author.
%
% (C) Laurens van der Maaten, 2010
% University California, San Diego / Delft University of Technology


    if length(unique(X(:))) == 2 && min(X(:)) == 0 && max(X(:)) == 1
        binary_data = true;
        mean_X = [];
        var_X = [];
    else
        binary_data = false;
        mean_X = mean(X, 1);
        X = bsxfun(@minus, X, mean_X);
        var_X = max(var(X, 1), eps);
        X = bsxfun(@rdivide, X, sqrt(var_X));        
    end

    % Pretrain the network
    origX = X;
    no_layers = length(layers);
    network = cell(1, no_layers);
    for i=1:no_layers

        % Print progress
        disp(['Training layer ' num2str(i) ' (size ' num2str(size(X, 2)) ' -> ' num2str(layers(i)) ')...']);
        
        if i ~= no_layers
          
            % Train layer using binary units
            if ~binary_data && i == 1
                network{i} = train_grbm(X, layers(i));
            else
                network{i} = train_rbm(X, layers(i));
            end                
                
            % Transform data using learned weights
            X = 1 ./ (1 + exp(-(bsxfun(@plus, X * network{i}.W, network{i}.bias_upW))));
        else
            
            % Train layer using Gaussian hidden units
            network{i} = train_lin_rbm(X, layers(i));
        end
    end
    
    % Perform backpropagation to minimize reconstruction error
    network = roll_out(network);
    network = backprop(network, origX, origX, 15);
    [foo, mappedX] = run_data_through_autoenc(network, origX);
    