function [clusters, labels] = em_mog(X, K, nrIter, init)
%EM_MOG Performs the EM algorithm for multivariate Gaussian mixtures
%
%   [clusters, labels] = em_mog(X, K, maxIter, init)
%
% Inputs:
%       X           dataset (matrix of size DxN)
%       K           number of clusters (default = 2)
%       maxIter     maximum number of iterations (default = 100)
%       init        initialization: 'random' or 'kmeans' (default = 'kmeans')
%
% Outputs:
%       clusters    structure:
%                       .centers: cluster centers
%                       .covs: cluster covariances
%                       .weights: mixture weights
%                       .L:  log-likelihood per iteration
%       labels      cluster assignment of data points
%
%
% (C) 2010, Laurens van der Maaten, UC San Diego


    % Default values
    if ~exist('K', 'var')
        K = 2;
    end
    if ~exist('nrIter', 'var')
        nrIter = 10;
    end
    
    % Initialize some variables
    [D, N] = size(X);
    clusters.L(1) = Inf;
    diff = Inf;
    iter = 0;
    min_std = 1e-4;

    % Assignment of cluster centers using k-means or random centers
    if ~exist('init', 'var') || strcmp(init, 'kmeans')
        [init, labels] = kmeans(X, K, 100);  
        clusters.centers = init.centers;
    else
        ind = randperm(N);
        clusters.centers = X(:,ind(1:K));
        DD = L2_distance(X, clusters.centers);
        [~, labels] = min(DD, [], 2);
    end
    
    % Compute covariances and mixture weights of clusters
    clusters.covs = zeros(D, D, K);
    clusters.weights = zeros(K, 1);
    for i=1:K
        ind = find(labels == i);
        clusters.covs(:,:,i) = X(:,ind) * X(:,ind)' + diag(repmat(min_std, [1 D]));
        clusters.weights(i) = length(ind) / N;
    end
    
    % Main loop
    while diff > 1e-7 && iter < nrIter
        
        % Update iteration number
        iter = iter + 1;
        
        % Expectation step
        % ===============================================================
        
        % Compute responsibilities (in log-domain)
        prob = zeros(K, N);
        for k=1:K            
            const = -(D ./ 2) .* log(2 * pi) - log(max(sqrt(max(0, det(clusters.covs(:,:,k)))), realmin));
            Xi = bsxfun(@minus, X, clusters.centers(:,k));
            prob(k,:) = log(max(clusters.weights(k), realmin)) + const - .5 .* sum((clusters.covs(:,:,k) \ Xi) .* Xi, 1);
        end
        shifts = max(prob, [], 1);
        resp = exp(bsxfun(@minus, prob, shifts));
        resp = bsxfun(@rdivide, resp, max(sum(resp, 1), realmin));
        [~, labels] = max(resp, [], 1);
        
        % Maximization step
        % ===============================================================
        
        % Compute new means, covariances, and mixture weights
        Nk = max(sum(resp, 2), realmin);
        for i=1:K
            clusters.centers(:,i) = (1 / Nk(i)) * sum(repmat(resp(i,:), [D 1]) .* X, 2);
            clusters.covs(:,:,i) = zeros(D, D);
            for j=1:N            
                clusters.covs(:,:,i) = clusters.covs(:,:,i) + ...
                            (resp(i, j) * ((X(:,j) - clusters.centers(:,i)) * (X(:,j) - clusters.centers(:,i))'));
            end
            clusters.covs(:,:,i) = (1 / Nk(i)) * (clusters.covs(:,:,i) + diag(repmat(min_std, [1 D])));
            clusters.weights(i) = Nk(i) / N;
        end
        
        % Log-likelihood computation
        % ===============================================================
        clusters.L(iter) = sum(bsxfun(@plus, log(sum(exp(bsxfun(@minus, prob, shifts)), 1)), shifts));
        if iter > 1, diff = clusters.L(iter) - clusters.L(iter - 1); end
        
        % Print iteration information
        fprintf('Iteration = %d, log-likelihood = %f\n', iter, clusters.L(iter));
    end    
