function [alpha, beta, rho, emission, gamma, eps] = forward_backward(X, model, emission) 
%FORWARD_BACKWARD Performs forward-backward algorithm in an HMM
%
%   [alpha, beta, rho, emission, gamma, eps] = forward_backward(X, model) 
%   [alpha, beta, rho, emission, gamma, eps] = forward_backward(X, model, emission)
%
% Performs the forward-backward algorithm on time series X in the HMM
% specified in model. The messages are returned in alpha and beta. The
% function also returns the normalization constants of the messages in rho,
% as well as the emission probabilities at all time steps.
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    % Initialize some variables
    N = size(X, 2);
    K = numel(model.pi);
    alpha = zeros(K, N);
    beta  = zeros(K, N);
    rho   = zeros(1, N);
    
    % Compute emission log-probabilities
    if ~exist('emission', 'var') || isempty(emission)
        eval(['emission = ' model.emission '_emission(X, model);']);
    end
    
%     % Perform forward-backward algorithm
%     [alpha, beta, rho] = forward_backward_emission(emission, model);
    
    % Compute message for first hidden variable
    alpha(:,1) = log(max(model.pi, realmin)) + emission(:,1);
    shift = max(alpha(:,1));
    alpha(:,1) = exp(alpha(:,1) - shift);
    rho(1) = sum(alpha(:,1));
    alpha(:,1) = alpha(:,1) ./ rho(1);
    rho(1) = log(max(rho(1), realmin)) + shift;
    
    % Perform forward pass
    for n=2:N
        alpha(:,n) = emission(:,n) + log(max(model.A' * alpha(:,n - 1), realmin));
        shift = max(alpha(:,n));
        alpha(:,n) = exp(alpha(:,n) - shift);
        rho(n) = sum(alpha(:,n));
        alpha(:,n) = alpha(:,n) ./ rho(n);
        rho(n) = log(max(rho(n), realmin)) + shift;
    end    
    
    % Perform backward pass
    if nargout > 1
        beta(:,N) = ones(K, 1);
        for n=N-1:-1:1
            tmp = log(max(beta(:,n + 1), realmin)) + emission(:,n + 1);
            shift = max(tmp);
            beta(:,n) = log(max(model.A * exp(tmp - shift), realmin));
            beta(:,n) = exp(beta(:,n) - rho(n + 1) + shift);
        end
    end
    
    % Compute posteriors over nodes
    if nargout > 4
        gamma = alpha .* beta;
        gamma = bsxfun(@rdivide, gamma, max(sum(gamma, 1), realmin));
    end
    
    % Compute posteriors over edges
    if nargout > 5
        if N > 1
            eps = repmat(realmin, [K K N - 1]);
            log_A = log(max(model.A, realmin));
            for n=2:N                 
                eps(:,:,n - 1) = log_A + repmat(log(max(alpha(:,n - 1), realmin)), [1 K]) + repmat(log(max(beta(:,n), realmin))' + emission(:,n)', [K 1]) - rho(n);
            end
            eps = sum(exp(eps - max(eps(:))), 3);
            eps = bsxfun(@rdivide, eps, max(sum(eps, 2), realmin));
        else
            eps = zeros(K, K);
        end
    end
 