function [C, dC] = hmm_grad(x, X, model, include_pi, include_A, include_mu, include_sigma, include_mix, rho, gamma, eps)
%HMM_GRAD Computes gradient of HMM log-likelihood for time series X
%
%   [C, dC] = hmm_grad(x, X, model, include_pi, include_A, include_mu, ...
%            include_sigma, include_mix, rho, gamma, eps)
%
% The function computes the gradient of the log-likelihood of the Hidden
% Markov Model specified by model and x, around time series X. 
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~exist('include_pi', 'var') || isempty(include_pi)
        include_pi = true;
    end
    if ~exist('include_A', 'var') || isempty(include_A)
        include_A = true;
    end
    if ~exist('include_mu', 'var') || isempty(include_mu)
        include_mu = true;
    end
    if ~exist('include_sigma', 'var') || isempty(include_sigma)
        include_sigma = true;
    end
    if ~exist('include_mix', 'var') || isempty(include_mix)
        include_mix = true;
    end

    % Convert current solution
    N = size(X, 2);
    K = numel(model.pi);
    D = size(model.mu, 1);
    ind = 1;
    if include_pi
        model.pi = reshape(x(ind:ind + numel(model.pi) - 1), size(model.pi)); ind = ind + numel(model.pi);
    end
    if include_A
        model.A  = reshape(x(ind:ind + numel(model.A)  - 1), size(model.A));  ind = ind + numel(model.A);
    end
    if include_mu
        model.mu = reshape(x(ind:ind + numel(model.mu) - 1), size(model.mu)); ind = ind + numel(model.mu);
    end
    if include_sigma && any(strcmpi(model.emission, {'gaussian', 'gaussian_mix'}))
        model.sigma = reshape(x(ind:ind + numel(model.sigma) - 1), size(model.sigma)); ind = ind + numel(model.sigma);
    end
    if include_mix && strcmpi(model.emission, 'gaussian_mix')
        model.mix = reshape(x(ind:ind + numel(model.mix)), size(model.mix));
    end
    
    % Compute negative log-likelihood
    if ~exist('rho', 'var') || ~exist('gamma', 'var') || ~exist('eps', 'var') || isempty(rho) || isempty(gamma) || isempty(eps)
        [~, ~, rho, ~, gamma, eps] = forward_backward(X, model);
    end
    C = -sum(rho);
    
    % Only compute gradient if requested
    if nargout > 1

        % Compute gradient with respect to state prior p(Z_1)
        if include_pi
            dPi = gamma(:,1) ./ max(model.pi, realmin);
        end
        
        % Compute gradient with respect to transition probabilities
        if include_A
            dA = eps ./ max(model.A, realmin);
        end
        
        % Compute gradient with respect to emission means and covariances
        if include_mu && strcmpi(model.emission, 'multinomial')
            dMu = sum(sum(bsxfun(@times, bsxfun(@times, gamma, X), model.mu)));
        else
            if include_mu || include_sigma 
                dMu = zeros(D, K);
                invSigma = zeros(size(model.sigma));
                for k=1:K
                    invSigma(:,:,k) = inv(model.sigma(:,:,k));
                    dMu(:,k) = sum(bsxfun(@times, gamma(k,:), invSigma(:,:,k) * bsxfun(@minus, X, model.mu(:,k))), 2);
                end
            end
        end
        
        % Compute gradient with respect to emission covariances
        if include_sigma && any(strcmpi(model.emission, {'gaussian', 'gaussian_mix'}))
            dSigma = zeros(D, D, K);
            for k=1:K
                Xi = bsxfun(@minus, X, model.mu(:,k));                
                dSigma(:,:,k) = -sum(gamma(k,:) / 2) .* invSigma(:,:,k) + ...
                                                   .5 * invSigma(:,:,k) * (bsxfun(@times, gamma(k,:), Xi) * Xi') * invSigma(:,:,k);
            end
        end
        % Does this also work for mixture models??? Needs extra resp's!!!
        
        % Compute gradient with respect to mixing proportions
        if include_mix && strcmpi(model.emission, 'gaussian_mix')
            dMix = zeros(size(model.mix));
        end
    
        % Convert gradient into its final form
        dC = [];
        if include_pi
            dC = [dC; -dPi(:)]; 
        end
        if include_A
            dC = [dC; -dA(:)];
        end
        if include_mu 
            dC = [dC; -dMu(:)];
        end
        if include_sigma && any(strcmpi(model.emission, {'gaussian', 'gaussian_mix'}))
            dC = [dC; -dSigma(:)];
        end
        if include_mix && strcmpi(model.emission, 'gaussian_mix')
            dC = [dC; -dMix(:)];
        end
    end
