function model = initialize_hmm(X, emis_type, initialization, K, scale, left_to_right, L)
%INITIALIZE_AAM Constructs an untrained Hidden Markov Model
%
%   model = initialize_hmm(X, emis_type, initialization, K, scale, left_to_right)
%   model = initialize_hmm(X, 'gaussian_mix', initialization, K, scale, left_to_right, L)
%
% The function constructs an untrained Hidden Markov Model with K states.
% The variable emis_type indicates the type of emission probabilities, and
% can be set to 'gaussian', 'gaussian_mix', or 'multinomial' (default =
% 'gaussian'). L is the number of mixture components for the Gaussian
% mixture emissions (default = 2). The variable D indicated the number of
% discrete states or the dimensionality of the continuous data. The
% variable initialization can be set to 'random' or 'pretrain' (default =
% 'random'). The pretrain setting initializes by training a mixture model.
% If left_to_right is set to true, the HMM cannot return to a previous 
% state (default = false). The variable scale sets the variance of the
% means, and only applies when initialization = 'random' (default = 0.1).
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~exist('emis_type', 'var') || isempty(emis_type)
        emis_type = 'gaussian';
    end
    if ~exist('initialization', 'var') || isempty(initialization)
        initialization = 'random';
    end
    if ~exist('left_to_right', 'var') || isempty(left_to_right)
        left_to_right = false;
    end
    if ~exist('L', 'var') || isempty(L)
        if strcmpi(emis_type, 'gaussian_mix'), L = 2; else L = 1; end
    end
    if ~exist('scale', 'var') || isempty(scale)
        scale = 0.1;
    end

    % Initialize transition probabilities
    D = size(X{1}, 1);
    if any(strcmpi(emis_type, {'gaussian', 'gaussian_mix'}))
        
        % Random initialization
        if strcmpi(initialization, 'random')
            model.mu = squeeze(randn(D, L, K)) * scale;
            model.sigma = squeeze(repmat(eye(D), [1 1 L K]));
            model.pi = repmat(1 / K, [K 1]);
            
        % Mixture of Gaussians initialization
        elseif strcmpi(initialization, 'pretrain')
            disp('Initializing using mixture of Gaussians...');
            mog = em_mog(cell2mat(X), K * L, 100, 'kmeans');
            model.mu = squeeze(reshape(mog.centers, [D L K]));
            model.sigma = squeeze(reshape(mog.covs, [D D L K]));
            if L > 1
                model.mix = reshape(mog.weights, [L K]);
                model.pi = repmat(1 / K, [K 1]);
            else
                model.pi = mog.weights;
            end
            clear mog
        else
            error('Unknown initialization method.');
        end
    else
        
        % Random initialization
        if strcmpi(initialization, 'random')
            model.mu = rand(D, K);
            
        % Mixture of multinomials initialization
        elseif strcmpi(initialization, 'pretrain')
            error('This is not yet implemented.');
        else
            error('Unknown initialization method.');
        end
    end
    
    % Initialize transition probabilities
    model.A = repmat(1 / K, [K K]);
    if left_to_right
        for k=1:K
            model.A(k, [1:k - 1 k + 2:end]) = 0;
        end
        model.A = bsxfun(@rdivide, model.A, sum(model.A, 2));
    end
    model.emission = emis_type;
