function [clusters, labels] = kmeans(X, K, nrIter, init)
%KMEANS Performs k-means clustering on dataset X
%
%   [clusters, labels] = kmeans(X, K, maxIter, init)
%
% Inputs:
%       X           dataset (matrix of size DxN)
%       K           number of clusters (default = 2)
%       maxIter     maximum number of iterations (default = 100)
%       init        initial cluster centers (optional)
%
% Outputs:
%       clusters    structure:
%                       .centers: cluster centers
%                       .changes: mean-square error per iteration
%                       .recerr:  reconstruction error per iteration
%       labels      cluster assignment of data points
%
%
% (C) 2007, Laurens van der Maaten, Maastricht University

    % Default values
    if ~exist('K', 'var')
        K = 2;
    end
    if ~exist('nrIter', 'var')
        nrIter = 100;
    end

    % Random assignment or predefined initial centers
    if ~exist('init', 'var')
        ind = randperm(size(X, 2));
        clusters.centers = X(:,ind(1:K));
    else
        clusters.centers = init;
    end
    
    % Initialize variables
    iter = 0;
    changes = Inf;
    old_labels = zeros(size(X, 2), 1);

    % Main loop
    while changes > 0 && iter < nrIter 

        % Update iteration number
        iter = iter + 1;
        err = 0;

        % Assign data points to centers (find nearest cluster center)
        D = L2_distance(X, clusters.centers);
        [distances, labels] = min(D, [], 2);

        % Compute new cluster centers (compute mean of assignments)
        for i=1:K    
            ind = find(labels == i);
            if ~isempty(ind)
                err = err + sum(distances(ind) .^ 2);
                clusters.centers(:,i) = sum(X(:,ind), 2) / length(ind);
            end
        end

        % Count number of changes and reconstruction error
        changes = sum(labels ~= old_labels);
        clusters.changes(iter) = changes;
        clusters.recerr(iter) = err;
        
        % Print iteration information
        fprintf('Iteration = %d, reconstruction error = %f, changed assignments = %d\n', iter, err, changes);

        % Prepare for next iteration
        old_labels = labels;
    end
    