function [C, dC] = mrf_grad(x, model, train_X, train_G, gamma, eta, log_prob, log_shift, eta_Z, log_Z)
%MRF_GRAD Log-likelihood + gradient of my simple MRF model


    % Decode current solution (empty x uses current model)
    if ~isempty(x)
        ind = 1;
        model.A  = reshape(x(ind:ind + numel(model.A)  - 1), size(model.A));  ind = ind + numel(model.A);
        if strcmpi(model.emission, 'multinomial')
            model.W = reshape(x(ind:ind + numel(model.W) - 1), size(model.W)); ind = ind + numel(model.W);

            % Compute multinomial parameters from weights
            model.mu = exp(model.W);
            model.mu = bsxfun(@rdivide, model.mu, sum(model.mu, 1));

        elseif strcmpi(model.emission, 'gaussian')
            model.mu = reshape(x(ind:ind + numel(model.mu) - 1), size(model.mu)); ind = ind + numel(model.mu);
            model.sigma = reshape(x(ind:ind + numel(model.sigma) - 1), size(model.sigma)); ind = ind + numel(model.sigma);
        else
            error('Unknown emission function.');
        end
    end
    
    % Initialize some variables
    N = length(train_X);
    K = size(model.A, 1);
    D = size(train_X{1}, 1);

    % Approximate beliefs and unnormalized log-probabilities for all objects
    if ~exist('gamma', 'var') && ~exist('eta', 'var') && ~exist('log_prob', 'var') && ~exist('log_shift', 'var')
        gamma     = cell(N, 1);             % beliefs over nodes
        eta       = cell(N, 1);             % beliefs over edges
        log_prob  = zeros(N, 1);            % log-partition function
        log_shift = zeros(N, 1);            % log-partition function
        for i=1:N
            [gamma{i}, eta{i}, log_prob(i), ~, ~, log_shift(i)] = sumprod_dai(train_X{i}, train_G{i}, model);
        end
    end
    
    % Approximate partition function
    if ~exist('eta_Z', 'var') && ~exist('log_Z', 'var')
        eta_Z = cell(N, 1);
        log_Z = zeros(N, 1);
        for i=1:N
            [~, eta_Z{i}, log_Z(i)] = sumprod_dai([], train_G{i}, model);
        end
    end
    
    % Compute negative log-likelihood    
    C = -sum(log_prob + log_shift - log_Z);
    
    % Compute gradients (if requested)
    if nargout > 1
        
        % Compute gradient w.r.t. transition log-probabilities
        dA = zeros(K, K);
        for i=1:N
            dA = dA + sum(eta{i}, 3) - sum(eta_Z{i}, 3);
        end
        dA = (dA + dA') / 2;            % force dA to be symmetric (is this necessary?)
        
        % Compute gradient w.r.t. multinomial parameters
        if strcmpi(model.emission, 'multinomial')
            dW = zeros(D, K);
            for i=1:N
                for k=1:K
                    dW(:,k) = dW(:,k) + sum(bsxfun(@times, gamma{i}(k,:), bsxfun(@minus, train_X{i}, model.mu(:,k))), 2);
                end
            end
            
        % Compute gradient w.r.t. Gaussian parameters
        elseif strcmpi(model.emission, 'gaussian')
            
            % Gradient w.r.t. means
            dMu = zeros(D, K);
            invSigma = zeros(size(model.sigma));
            for k=1:K
                invSigma(:,:,k) = inv(model.sigma(:,:,k));
            end
            for i=1:N
                for k=1:K                    
                    dMu(:,k) = dMu(:,k) + sum(bsxfun(@times, gamma{i}(k,:), invSigma(:,:,k) * bsxfun(@minus, train_X{i}, model.mu(:,k))), 2);
                end
            end
            
            % Gradient w.r.t. covariances
            dSigma = zeros(D, D, K);
            for i=1:N
                for k=1:K
                    Xi = bsxfun(@minus, train_X{i}, model.mu(:,k));                
                    dSigma(:,:,k) = dSigma(:,:,k) - sum(gamma{i}(k,:) / 2) .* invSigma(:,:,k) + ...
                                                                        .5 .* invSigma(:,:,k) * (bsxfun(@times, gamma{i}(k,:), Xi) * Xi') * invSigma(:,:,k);
                end
            end
        else
            error('Unknown emission function.');
        end
        
        % Encode gradients
        if strcmpi(model.emission, 'multinomial')
            dC = -[dA(:); dW(:)];
        elseif strcmpi(model.emission, 'gaussian')
            dC = -[dA(:); dMu(:); dSigma(:)];
        else
            error('Unknown emission function.');
        end    
    end
    