function err = mutagenicity_experiment(representation, classifier, states)
%MUTAGENICITY_EXPERIMENT Perform experiment on mutagenicity data set
%
%   err = mutagenicity_experiment(representation, classifier, states)
%
% The function performs a Fisher kernel learning experiment on the 
% mutagenicity data set. Possible values for representation are
% 'mrf_posterior', 'mrf_gradient', and 'fkl'. Possible values for
% classifier are 'bayes' (posteriors only), 'softmax', 'svm', and 'lmnn'.
% You can specify the number of HMM states for which to run the experiments
% through states.
% The function returns a bunch of generalization errors in a cell array
% with one cell per number of states.
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~exist('states', 'var') || isempty(states)
        states = 10;
    end
    if ~exist('classifier', 'var') || isempty(classifier)
        classifier = 'all';
    end    
    if ischar(classifier) 
        if strcmpi(classifier, 'all')
            classifier = {'bayes', 'softmax', 'svm', 'lmnn'};
        else
            classifier = {classifier};
        end
    end

    % Load and preprocess data
    warning('This experiments requires a version of libDAI that is compiled for your platform!');
    disp('Preprocessing data...');
    addpath('libsvm-2.9');
    rmpath([matlabroot '/toolbox/stats']);
    addpath('prtools'); 
    addpath(genpath('drtoolbox'));
    X = []; labels = [];
    load 'data/mutagenicity.mat'
    count = 0;
    for i=1:length(X)
        count = count + size(X{i}, 1);
    end
    list_X = cell(count, 1);
    count = 1;
    for i=1:length(X)
        list_X(count:count + size(X{i}, 1) - 1) = X{i};
        count = count + length(X{i});
    end
    voc = unique(list_X);                   % list of all possible atoms
    no_atoms = length(voc);
    num_X = cell(length(X), 1);
    for i=1:length(X)
        num_X{i} = zeros(no_atoms, size(X{i}, 1));
        for j=1:length(voc)
            num_X{i}(j, ismember(X{i}, voc{j})) = 1;
        end
    end
    X = num_X;
    no_labels = length(unique(T));
    clear num_X count i j list_X num_X
    
    % Split data into training and test set
    no_folds = 10;
    fold_size = floor(length(X) ./ no_folds);
    perm = randperm(length(X)); fold_ind = 1;
    train_ind = perm([1:fold_ind - 1 fold_ind + fold_size:end]);
    test_ind  = perm(fold_ind:fold_ind + fold_size - 1);
    train_X = X(train_ind);
    train_G = G(train_ind);
    train_labels = T(train_ind);
    test_X = X(test_ind);
    test_G = G(test_ind);
    test_labels = T(test_ind);
%     fold_ind = fold_ind + fold_size;
    
    % Initialize some variables
    err = ones(length(states), length(classifier));
    
    % Repeat this experiment for various number of states
    for index=1:length(states)
        
        % Construct time series representation
        switch representation

            case 'mrf_posterior'
    
                % Train the MRF models                
                model = cell(no_labels, 1);
                for i=1:no_labels
                    model{i} = train_mrf(train_X(train_labels == i), train_G(train_labels == i), states(index), 'multinomial');
                end
                
                % Compute the data log-likelihoods
                train_feat = zeros(length(train_X), no_labels);                    
                test_feat  = zeros(length(test_X),  no_labels);
                for i=1:length(train_X)
                    for j=1:length(model)
                        train_feat(i, j) = mrf_grad([], model{j}, train_X(i), train_G(i));
                    end
                end                    
                for i=1:length(test_X)
                    for j=1:length(model)
                        test_feat(i, j) = mrf_grad([], model{j}, test_X(i), test_G(i));
                    end
                end
                
            case 'mrf_gradient'
                
                % Train a single MRF on all data
                model = train_mrf(train_X, train_G, states(index), 'multinomial');
                
                % Compute gradient representations
                train_feat = zeros(length(train_X), states(index) * states(index) + no_atoms * states(index));
                test_feat  = zeros(length(test_X),  states(index) * states(index) + no_atoms * states(index));
                for i=1:length(train_X)
                    [~, tmp] = mrf_grad([], model, train_X(i), train_G(i));
                    train_feat(i,:) = tmp';
                end        
                for i=1:length(test_X)
                    [~, tmp] = mrf_grad([], model, test_X(i), test_G(i));
                    test_feat(i,:) = tmp;
                end                
                
            case 'fkl'
                
                % Perform Fisher kernel learning
                lambda = 0;
                [train_feat, model, M] = stse_mrf_metric(train_X, train_G, train_labels, states(index), lambda);
                train_feat = train_feat'; 
                M = M';
                
                % Compute features for test data
                test_feat  = zeros(length(test_X), states(index) * states(index) + no_atoms * states(index));               
                gamma = cell(length(test_X), 1);
                eta   = cell(length(test_X), 1);
                eta_Z = cell(length(test_X), 1);
                log_prob  = zeros(length(test_X), 1);
                log_shift = zeros(length(test_X), 1);
                log_Z     = zeros(length(test_X), 1);
                for n=1:length(test_X)
                    [gamma{n}, eta{n}, log_prob(n), ~, ~, log_shift(n)] = sumprod_dai(test_X{n}, test_G{n}, model);
                    [~, eta_Z{n}, log_Z(n)] = sumprod_dai([], G{n}, model);
                end
                for n=1:length(test_X)
                    [~, tmp] = mrf_mu_grad([model.A(:); model.mu(:)], model, test_X(n), test_G(n), gamma(n), eta(n), log_prob(n), log_shift(n), eta_Z(n), log_Z(n));
                    test_feat(n,:) = tmp';
                end
                
                % Apply metric on both representations
                train_feat = bsxfun(@times, train_feat, sqrt(M));
                test_feat  = bsxfun(@times,  test_feat, sqrt(M));
                
            otherwise
                error('Unknown representation.');
        end
        
        % Perform classification
        save(['features_' representation '_' num2str(states(index)) '.mat']);
        if ischar(classifier)
            classifier = {classifier};
        end
        for i=1:length(classifier)
            
            % Select current classifier
            classif = classifier{i};
            switch classif
                case 'bayes'

                    if ~strcmpi(representation, 'mrf_posterior')
                        warning('The Bayes classifier can only be used in combination with the HMM posterior representation. Returning error of 1.');
                        est_labels = test_labels + 1;
                    else

                        % Multiply with class prior
                        prior = zeros(1, no_labels);
                        for j=1:no_labels
                            prior(j) = sum(train_labels == j);
                        end
                        prior = prior ./ sum(prior);
                        test_feat = bsxfun(@plus, test_feat, log(prior));

                        % Perform MAP classification
                        [~, est_labels] = max(test_feat, [], 2);
                    end

                case 'softmax'

                    % Cross-validation to determine value of L2-regularizer
                    N = length(train_X);
                    test_feat  = test_feat  - min(train_feat(:));
                    train_feat = train_feat - min(train_feat(:));
                    test_feat  = test_feat  ./ (max(train_feat(:)));
                    train_feat = train_feat ./ (max(train_feat(:)));
                    lambda = [0 0.0001 0.0002 0.0005 0.001 0.002 0.005 0.01 0.02 0.05 0.1 0.2 0.5 1 2 5];
                    lambda_err = nan(length(lambda), 1);
                    valid_ind = randperm(N)';
                    trn_valid_ind = valid_ind(1:round(.9 * N));
                    tst_valid_ind = valid_ind(1+round(.9 * N):end);                    
                    for j=1:length(lambda)
                        disp(['Training logistic regressor ' num2str(j) ' of ' num2str(length(lambda)) '...']);
                        [E, E_bias] = train_logreg(train_feat(trn_valid_ind,:)', train_labels(trn_valid_ind), lambda(j));
                        est_labels = eval_logreg(train_feat(tst_valid_ind,:)', E, E_bias)';
                        lambda_err(j) = sum(est_labels ~= train_labels(tst_valid_ind)) ./ numel(tst_valid_ind);
                    end
                    [~, min_ind] = min(lambda_err);
                    lambda = lambda(min_ind);
                    disp(['Using value of lambda of ' num2str(lambda) '...']);                    

                    % Train and test a softmax on the posterior probabilities
                    [E, E_bias] = train_logreg(train_feat', train_labels, lambda);
                    est_labels = eval_logreg(test_feat', E, E_bias)';

                case 'svm'

                    % Normalize data
                    disp('Performing cross-validation over slack variable...');
                    N = length(train_X);
                    test_feat  = test_feat  - min(train_feat(:));
                    train_feat = train_feat - min(train_feat(:));
                    test_feat  = test_feat  ./ (max(train_feat(:)));
                    train_feat = train_feat ./ (max(train_feat(:)));
                    
                    % Cross-validation to determine value of slack variable
                    C = 0.25:.25:2.5;
                    C_err = nan(length(C), 1);
                    valid_ind = randperm(N)';
                    trn_valid_ind = valid_ind(1:round(.9 * N));
                    tst_valid_ind = valid_ind(1+round(.9 * N):end);                    
                    trn_valid_feat = train_feat(trn_valid_ind,:);
                    tst_valid_feat = train_feat(tst_valid_ind,:);
                    train_K = dataset(trn_valid_feat * trn_valid_feat', train_labels(trn_valid_ind));
                    test_K  = dataset(tst_valid_feat * trn_valid_feat', train_labels(tst_valid_ind));
%                     [train_K, rem_ind] = remove_duplicates(train_K, 1e-5);      % removes duplicate points
%                     train_K(:,rem_ind) = [];
%                     test_K(:,rem_ind) = [];
                    for j=1:length(C)
                        disp(['Training SVM ' num2str(j) ' of ' num2str(length(C)) '...']);
                        svm = libsvc(train_K, 0, C(j));
                        est_labels = labeld(map(test_K, svm));
                        C_err(j) = sum(est_labels ~= train_labels(tst_valid_ind)) ./ numel(tst_valid_ind);
                    end
                    [~, min_ind] = min(C_err);
                    C = C(min_ind);
                    disp(['Using value of C of ' num2str(C) '...']);

                    % Compute (practical) Fisher kernel
                    disp(['Performing final SVM training (slack = ' num2str(C) ')...']);
                    train_K = dataset(train_feat * train_feat', train_labels);
                    test_K  = dataset(test_feat  * train_feat',  test_labels);
%                     [train_K, rem_ind] = remove_duplicates(train_K, 1e-5);      % removes duplicate points
%                     train_K(:,rem_ind) = [];
%                     test_K(:,rem_ind) = [];
                    
                    % Train and test final Fisher kernel SVM
                    svm = libsvc(train_K, 0, C);
                    est_labels = labeld(map(test_K, svm));

                case 'lmnn'

                    % Perform PCA + LMNN
                    addpath(genpath('mLMNN'));
                    [train_feat2, mapping] = compute_mapping(train_feat, 'PCA', min(min(size(train_feat)), 200)); 
                    metric = lmnn2(train_feat2', train_labels', 5, 'quiet', 1, 'maxiter', 500, 'validation', 0.1);

                    % Perform classification of test data 
                    test_feat2 = out_of_sample(test_feat, mapping);
                    [~, est_labels] = energyclassify(metric, train_feat2', train_labels', test_feat2', test_labels', 9); % should cross-validate over k?
                    est_labels = est_labels';
                    
                case '1nn'          % assumes we have distance matrix D

                    % Perform 1-NN classifications
                    [~, ii] = sort(D, 1, 'ascend');
                    est_labels = train_labels(ii(1,:));

                otherwise
                    error('Unknown classifier.');                    
            end

            % Measure generalization error on test set
            err(index, i) = sum(est_labels ~= test_labels) ./ numel(test_labels);
            disp(['Generalization error on test set (' classif ' classifier): ' num2str(err(index, i))]); 
        end
    end    
    