function [U, model, mean_X, var_X, err] = stse(X, labels, no_states, sigma, lambda, include_pi, include_A, include_mu)
%STSE Performs supervised time series embedding
%
%   [U, model, mean_X, var_X, err] = stse(X, labels, no_states, sigma, lambda, include_pi, include_A, include_mu)   
%
% The function performs supervised time series embedding (STSE). STSE trains 
% an HMM in such a way that time series with the same class are modeled 
% close together. 
% The variable sigma is the scale variable for Gaussian similarities
% (default = 0.25). The variable lambda determines the amount of L2-regula-
% rization on the HMM gradients (default  = 0).
% The function returns the time series embedding for the data X in U, and 
% the STSE model in model.
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~exist('sigma', 'var') || isempty(sigma)
        sigma = 0.25;
    end
    if ~exist('lambda', 'var') || isempty(lambda)
        lambda = 0;
    end
    if ~exist('include_pi', 'var') || isempty(include_pi)
        include_pi = true;
    end
    if ~exist('include_A', 'var') || isempty(include_A)    
        include_A  = true;
    end
    if ~exist('include_mu', 'var') || isempty(include_mu)    
        include_mu = true;
    end
    if size(X, 2) == 1
        X = X';
    end

    % Initialize some variables
    N = length(X);
    K = no_states;
    D = size(X{1}, 1);
    use_log = false;
    max_iter = 25;
    C = nan(max_iter, 1);
    err = nan(max_iter, 1);
    total_length = 0;
    for i=1:length(X)    
        total_length = total_length + size(X{i}, 2);
    end
    rho   = cell(N, 1);    
    gamma = cell(N, 1);
    eps   = cell(N, 1);
        
    % Make frames zero-mean, unit-variance
    dims = cell(size(X)); for i=1:length(dims), dims{i} = 2; end
    mean_X = sum(cell2mat((cellfun(@sum, X, dims, 'UniformOutput', false))), 2) ./ total_length;
    var_X = zeros(size(X{i}, 1), 1);
    for i=1:length(X)
       var_X = var_X + sum(bsxfun(@minus, X{i}, mean_X) .^ 2, 2);
    end
    var_X = var_X / total_length;
    for i=1:length(X)
       X{i} = bsxfun(@rdivide, bsxfun(@minus, X{i}, mean_X), sqrt(var_X));
    end
    
    % Initialize random HMM
    model = initialize_hmm(X, 'gaussian', 'random', no_states, 0.1, false);
    
    % Initialize scale normalization for STSE
    U = zeros(include_pi * K + include_A * K * K + include_mu * D * K, N);
    for n=1:N
        [~, U(:,n)] = hmm_grad([model.pi(:); model.A(:); model.mu(:); model.sigma(:)], X{n}, model, include_pi, include_A, include_mu, false, false);
    end
    sum_U = sum(U .^ 2, 1);
    sigma = sigma * max(max(bsxfun(@plus, sum_U, bsxfun(@plus, sum_U', -2 * (U' * U)))));
    
    % Compute binary same-class indicator matrix
    same_class = repmat(false, [N N]);
    for n=1:N
        same_class(n,:) = (labels == labels(n));
    end
    same_class(1:N + 1:end) = false;
            
    % Encode solution
    x = [];
    if include_pi, x = [x; model.pi(:)]; end
    if include_A,  x = [x; model.A(:)];  end
    if include_mu, x = [x; model.mu(:)]; end
%     x = [x; sigma];
    
    % Perform alternating optimization (= EM algorithm)
    iter = 0;
    models = cell(max_iter, 1);
%     sigmas = nan(max_iter, 1);
    while iter < max_iter && (iter < 10 || (C(iter) - C(iter - 1) < -1e-4 && C(iter - 1) - C(iter - 2) < -1e-4))
        
%         % Plot Gaussian means if data is 2D
%         if size(X{1}, 1) == 2
%             scatter(XX(1:5000, 1), XX(1:5000, 2)); hold on
%             for k=1:no_states
%                 plot_gauss(model.mu(:,k), model.sigma(:,:,k)); hold on
%             end
%             hold off; drawnow
%         end
    
        % Compute responsibilities (E-step)
        iter = iter + 1;
        for n=1:N
            [~, ~, rho{n}, ~, gamma{n}, eps{n}] = forward_backward(X{n}, model);
        end
        
        % Compute gradient
        [C(iter), dC, U] = stse_grad(x, X, model, rho, gamma, eps, same_class, use_log, sigma, lambda, include_pi, include_A, include_mu);
        disp(['Iteration ' num2str(iter) ': cost function is ' num2str(C(iter))]);
        
        % Compute nearest neighbor error
        err(iter) = nn_error(U', labels);
        disp(['Nearest neighbor error: ' num2str(err(iter))]);
        figure(1), plot(1:iter, err(1:iter));
        figure(2), plot(1:iter, C(1:iter));
        drawnow

        % Perform gradient step given responsibilities (M-step)
        eta = 1e-4;
        x = x + eta * dC;
%         checkgrad('stse_grad', x, 1e-5, X, model, rho, gamma, eps, same_class, use_log, sigma, lambda, include_pi, include_A, include_mu)
%         opts.LS = 2;
%         opts.LS_init = 1;
%         opts.MaxIter = 1;
%         opts.Method = 'sd';
%         opts.Display = 'final';
%         x = minFunc(@stse_grad, x, opts, X, model, rho, gamma, eps, same_class, use_log, sigma, lambda, include_pi, include_A, include_mu);        
        
        % Decode solution
        iii = 1;
        if include_pi, model.pi = reshape(x(iii:iii + numel(model.pi) - 1), size(model.pi)); iii = iii + numel(model.pi); end
        if include_A,  model.A  = reshape(x(iii:iii + numel(model.A)  - 1), size(model.A));  iii = iii + numel(model.A);  end
        if include_mu, model.mu = reshape(x(iii:iii + numel(model.mu) - 1), size(model.mu)); end
        models{iter} = model;
%         sigmas(iter) = x(end);
%         disp(['Current value of sigma: ' num2str(sigmas(iter))]);

        % Project onto manifold of valid HMMs
        model.pi = model.pi ./ sum(model.pi);
        model.A = bsxfun(@rdivide, model.A, sum(model.A, 2));
    end
    
    % Use the best model as final model
    [~, min_ind] = min(err);
    model = models{min_ind};
    x = [];
    if include_pi, x = [x; model.pi(:)]; end
    if include_A,  x = [x; model.A(:)];  end
    if include_mu, x = [x; model.mu(:)]; end
%     x = [x; sigmas(min_ind)];
    for n=1:N
        [~, ~, rho{n}, ~, gamma{n}, eps{n}] = forward_backward(X{n}, model);
    end
    [C(end), ~, U] = stse_grad(x, X, model, alpha, beta, rho, emission, same_class, use_log, sigma, lambda, include_pi, include_A, include_mu);        
