function [C, dC, grad] = stse_grad(x, X, model, rho, gamma, eps, same_class, use_log, sigma, lambda, include_pi, include_A, include_mu)
%STSE_GRAD Gradient of supervised time series embedding
%
%   [C, dC, U] = stse_grad(x, X, model, rho, gamma, eps, ...
%                                      same_class, use_log, sigma, lambda, ...
%                                      include_pi, include_A, include_mu)
%
% Computes the gradient of the function that aims to maximize the variance
% of the gradient of an HMM with respect to the means of the Gaussians. The
% variable lambda (0 <= lambda <= 1) weighs the two variance terms.
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    % Initialize some variables
    assert(strcmpi(model.emission, 'gaussian'));
    N = length(X);
    [D, K] = size(model.mu);
    assert(length(x) == include_pi * K + include_A * K * K + include_mu * D * K);
%     assert(length(x) == include_pi * K + include_A * K * K + include_mu * D * K + 1);
        
    % Convert current solution
    ind = 1;
    if include_pi, model.pi = reshape(x(ind:ind + numel(model.pi) - 1), size(model.pi)); ind = ind + numel(model.pi); end
    if include_A,  model.A  = reshape(x(ind:ind + numel(model.A)  - 1), size(model.A));  ind = ind + numel(model.A);  end
    if include_mu, model.mu = reshape(x(ind:ind + numel(model.mu) - 1), size(model.mu)); end
    include_sigma = false; 
    include_mix = false;
%     sigma = x(end);
    
    % Compute maximum likelihood gradient w.r.t. HMM parameters (isotropic cov.)
    grad = zeros(length(x), N);
%     grad = zeros(length(x) - 1, N);
    for n=1:N
        [~, grad(:,n)] = hmm_grad([model.pi(:); model.A(:); model.mu(:)], X{n}, model, include_pi, include_A, include_mu, include_sigma, include_mix, rho{n}, gamma{n}, eps{n});
%         checkgrad('hmm_grad', [model.pi(:); model.A(:); model.mu(:)], 1e-5, X{n}, model, include_pi, include_A, include_mu, include_sigma, include_mix)
    end
    
    % Compute pairwise affinities between gradients
    sum_grad = sum(grad .^ 2, 1);
    distances = bsxfun(@plus, sum_grad, bsxfun(@plus, sum_grad', -2 * (grad' * grad)));
%     disp(['Mean: ' num2str(mean(distances(:))) ', max: ' num2str(max(distances(:))) ', sigma: ' num2str(sigma)]);
    P = exp(-distances ./ sigma);
    P(1:N + 1:end) = 0;
    P = max(P, realmin);
    P = bsxfun(@rdivide, P, sum(P, 2));
    
    % Compute value of cost function
    if use_log
        Pi = sum(same_class, 1);
        C = -sum(sum(log(P(same_class)))) + lambda * sum(grad(:));
    else
        Pi = zeros(1, N);
        for n=1:N
            Pi(n) = sum(P(n, same_class(:,n)));
        end
        C = -sum(Pi) + lambda * sum(grad(:) .^ 2);
    end
    
    % Only compute gradient if requested
    if nargout > 1
        
        % Compute gradient of gradient w.r.t. HMM prior state probabilities
        ind = 1;
%         grad_grad = zeros(length(x) - 1, N);
        grad_grad = zeros(length(x), N);
        if include_pi
            for n=1:N
                grad_grad(ind:ind + K - 1, n) = -gamma{n}(:,1) ./ (model.pi .^ 2);
            end
            ind = ind + K;
        end
        
        % Compute gradient of gradient w.r.t. HMM transitions
        if include_A
            for n=1:N
                grad_grad(ind:ind + K * K - 1, n) = -reshape(eps{n} ./ (model.A .^ 2), [K * K 1]);
            end
            ind = ind + K * K;
        end
        
        % Compute gradient of gradient w.r.t. HMM means
        if include_mu
            for n=1:N
                for k=1:K
                    grad_grad(ind + (k - 1) * D:ind + k * D - 1, n) = -sum(gamma{n}(k,:));
                end
            end     
        end

        % Compute gradient of NCA cost function w.r.t. all HMM parameters
        dC = zeros(length(x), 1);
%         dC = zeros(length(x) - 1, 1);
        for n=1:N
            tmp = bsxfun(@minus, grad(:,n), grad) .* ...
                  bsxfun(@minus, grad_grad(:,n), grad_grad);
            if use_log
                dC = dC + sum(tmp(:,same_class(:,n)), 2) ...
                        - Pi(n) .* sum(bsxfun(@times, P(n,:), tmp), 2);
            else
                tmp = bsxfun(@times, P(n,:), tmp);
                dC = dC + sum(tmp(:,same_class(:,n)), 2) ...
                        - Pi(n) .* sum(tmp, 2);
            end
        end
        dC = (-2 / sigma) * dC(:) - lambda * 2 * sum(grad .* grad_grad, 2);
        
%         % Compute gradient of cost function w.r.t. sigma
%         distances = distances ./ (sigma .^ 2);
%         if use_log
%             dC = [dC; -sum(distances(same_class)) + sum(Pi' .* sum(P .* distances, 2))];
%         else
%             distances = P .* distances;
%             dC = [dC; -sum(distances(same_class)) + sum(Pi' .* sum(distances, 2))];
%         end            
    end
    