function [U, model, M] = stse_mrf_metric(X, G, labels, no_states, lambda)
%STSE_MRF_METRIC Performs supervised time series embedding
%
%   [U, model, M] = stse_mrf_metric(X, G, labels, no_states, lambda)   
%
% The function performs supervised time series embedding (STSE). STSE trains 
% an HMM in such a way that time series with the same class are modeled 
% close together. 
% The variable sigma is the scale variable for Gaussian similarities
% (default = 0.25). The variable lambda determines the amount of L2-regula-
% rization on the HMM gradients (default  = 0).
% The function returns the time series embedding for the data X in U, and 
% the STSE model in model.
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~exist('lambda', 'var') || isempty(lambda)
        lambda = 0;
    end
    if size(X, 2) == 1
        X = X';
    end
    addpath(genpath('minFunc'))

    % Initialize some variables
    N = length(X);
    K = no_states;
    D = size(X{1}, 1);
    use_log = false;
    max_iter = 20;
    C   = nan(max_iter, 1);
    err = nan(max_iter, 1);
    Ms = cell(max_iter, 1);
    models = cell(max_iter, 1);
    total_length = 0;
    for i=1:length(X)    
        total_length = total_length + size(X{i}, 2);
    end
    
    % Initialize random MRF
    model.emission = 'multinomial';
    model.A  = randn(K, K) * .1;
    model.A  = model.A + model.A';
    model.mu = rand(D, K);
    model.mu = bsxfun(@rdivide, model.mu, sum(model.mu, 1));
    model.W  = log(max(model.mu, realmin));
    
    % Initialize feature weighting for STSE
    U = zeros(K * K + D * K, N);
    for n=1:N
        [~, U(:,n)] = mrf_mu_grad([model.A(:); model.mu(:)], model, X(n), G(n));
    end
    M = min(max(1 ./ mean(U .^ 2, 2), 1e-7), 1e7);
    
    % Compute binary same-class indicator matrix
    same_class = repmat(false, [N N]);
    for n=1:N
        same_class(n,:) = (labels == labels(n));
    end
    same_class(1:N + 1:end) = false;
    
    % Initialize some variables
    gamma     = cell(N, 1);
    eta       = cell(N, 1);
    eta_Z     = cell(N, 1);
    log_prob  = zeros(N, 1);
    log_shift = zeros(N, 1);
    log_Z     = zeros(N, 1);
            
    % Encode solution
    x = [model.A(:); model.mu(:); M(:)];
    
    % Perform alternating optimization
    iter = 0;
    while iter < max_iter && (iter < 10 || (C(iter) - C(iter - 1) < -1e-4 && C(iter - 1) - C(iter - 2) < -1e-4))
        
        % Compute beliefs and log-partition function (E-step)
        iter = iter + 1;
        model.W = log(max(model.mu, realmin));
        for n=1:N
            [gamma{n}, eta{n}, log_prob(n), ~, ~, log_shift(n)] = sumprod_dai(X{n}, G{n}, model);
            [~, eta_Z{n}, log_Z(n)] = sumprod_dai([], G{n}, model);
        end
        
        % Compute gradient
        [C(iter), ~, U] = stse_mrf_metric_grad(x, X, G, model, gamma, eta, log_prob, log_shift, eta_Z, log_Z, same_class, use_log, lambda);
        disp(['Iteration ' num2str(iter) ': cost function is ' num2str(C(iter))]);
        if iter > 2 && C(iter - 1) - C(iter) < 1e-5
            break;
        end
        
        % Compute nearest neighbor error
        err(iter) = nn_error(U', labels, M);
        disp(['Nearest neighbor error: ' num2str(err(iter))]);
%         figure(1), plot(1:iter, err(1:iter));
%         figure(2), plot(1:iter, C(1:iter));

        % Perform gradient step given responsibilities (M-step)
%         checkgrad('stse_mrf_metric_grad', x, 1e-7, X, G, model, gamma, eta, log_prob, log_shift, eta_Z, log_Z, same_class, use_log, lambda)
        opts.LS = 2;
        opts.LS_init = 1;
        opts.MaxIter = 5;
        opts.Method = 'sd';
        opts.Display = 'final';
        x = minFunc(@stse_mrf_metric_grad, x, opts, X, G, model, gamma, eta, log_prob, log_shift, eta_Z, log_Z, same_class, use_log, lambda);        
        
        % Decode solution
        iii = 1;
        model.A  = reshape(x(iii:iii + numel(model.A)  - 1), size(model.A));  iii = iii + numel(model.A);
        model.mu = reshape(x(iii:iii + numel(model.mu) - 1), size(model.mu)); iii = iii + numel(model.mu);
        M = x(iii:end);
        
        % Project back onto manifold of valid models
        M = min(max(M, 1e-7), 1e7);
        model.mu = max(realmin, model.mu);
        model.mu = bsxfun(@rdivide, model.mu, sum(model.mu, 1));
        Ms{iter} = M;
        models{iter} = model;
    end
    
    % Use the best model as final model
    [~, min_ind] = min(err);
    model = models{min_ind};
    M = Ms{min_ind};
    
    % Compute final gradient
    for n=1:N
        [gamma{n}, eta{n}, log_prob(n), ~, ~, log_shift(n)] = sumprod_dai(X{n}, G{n}, model);
        [~, eta_Z{n}, log_Z(n)] = sumprod_dai([], G{n}, model);
    end
    [~, ~, U] = stse_mrf_metric_grad([model.A(:); model.mu(:); M(:)], X, G, model, gamma, eta, log_prob, log_shift, eta_Z, log_Z, same_class, use_log, lambda);    
