function [C, dC, grad] = stse_mrf_metric_grad(x, X, G, model, gamma, eta, log_prob, log_shift, eta_Z, log_Z, same_class, use_log, lambda)
%STSE_METRIC_GRAD Gradient of supervised time series embedding for MRFs
%
%   [C, dC, U] = stse_metric_grad(x, X, G, model, gamma, eta, eta_Z, log_Z, ...
%                                      same_class, use_log, lambda)
%
% Computes the gradient of the function that aims to maximize the variance
% of the gradient of a binary MRF with discrete variables with respect to 
% the MRF parameters and metric. The variable lambda (0 <= lambda <= 1) 
% weighs the two variance terms.
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~strcmpi(model.emission, 'multinomial')
        error('This function is only implemented for multinomial emission distributions.');
    end

    % Initialize some variables
    N = length(X);
    [D, K] = size(model.mu);
    assert(length(x) == 2 * (K * K + D * K));
        
    % Convert current solution
    ind = 1;
    model.A  = reshape(x(ind:ind + numel(model.A)  - 1), size(model.A));  ind = ind + numel(model.A);
    model.mu = reshape(x(ind:ind + numel(model.mu) - 1), size(model.mu)); ind = ind + numel(model.mu);
    model.W  = log(max(model.mu, realmin));
    M = x(ind:end);             % metric
    x = x(1:ind - 1);           % MRF parameters
        
    % Compute maximum likelihood gradient w.r.t. MRF parameters
    grad = zeros(length(x), N);
    for n=1:N
        [~, grad(:,n)] = mrf_mu_grad([model.A(:); model.mu(:)], model, X(n), G(n), gamma(n), eta(n), log_prob(n), log_shift(n), eta_Z(n), log_Z(n));            
    end
    
    % Compute pairwise affinities between gradients under metric M
    distances = zeros(N, N);
    for n=1:N
        diff = bsxfun(@minus, grad(:,n), grad);
        distances(:,n) = sum(bsxfun(@times, diff, M) .* diff, 1)';
    end
    P = exp(-distances);
    P(1:N + 1:end) = 0;
    P = max(P, realmin);
    P = bsxfun(@rdivide, P, sum(P, 2));
    
    % Compute value of cost function
    if use_log
        Pi = sum(same_class, 1);
        C = -sum(sum(log(P(same_class)))) + lambda * sum(grad(:));
    else
        Pi = zeros(1, N);
        for n=1:N
            Pi(n) = sum(P(n, same_class(:,n)));
        end
        C = -sum(Pi) + lambda * sum(grad(:) .^ 2);
    end
    
    % Only compute gradient if requested
    if nargout > 1
        
        % Compute gradient of gradient w.r.t. MRF transitions
        ind = 1;
        grad_grad = zeros(length(x), N);        
        ind = ind + K * K;
        
        % Compute gradient of gradient w.r.t. MRF emissions
        for k=1:K
            for n=1:N
                tmp = -bsxfun(@rdivide, sum(bsxfun(@times, gamma{n}(k,:), X{n}), 2), max(model.mu(:,k) .^ 2, realmin));
                grad_grad(ind:ind + D - 1, n) = tmp(:);
            end
            ind = ind + D;
        end
        
        % Compute gradient of NCA cost function w.r.t. all MRF parameters
        dC = zeros(length(x), 1);
        for n=1:N
            tmp = bsxfun(@minus, grad(:,n), grad) .* ...
                  bsxfun(@minus, grad_grad(:,n), grad_grad);
            if use_log
                dC = dC + sum(tmp(:,same_class(:,n)), 2) ...
                        - Pi(n) .* sum(bsxfun(@times, P(n,:), tmp), 2);
            else
                tmp = bsxfun(@times, P(n,:), tmp);
                dC = dC + sum(tmp(:,same_class(:,n)), 2) ...
                        - Pi(n) .* sum(tmp, 2);
            end
        end
        dC = -2 .* (M .* dC) + lambda * 2 * sum(grad .* grad_grad, 2);
        
        % Compute gradient w.r.t. metric
        dCdM = zeros(numel(M), 1);
        for n=1:N
            tmp = bsxfun(@minus, grad(:,n), grad) .^ 2;
            tmp = bsxfun(@times, P(n,:), tmp);
            dCdM = dCdM + sum(tmp(:,same_class(:,n)), 2) - (Pi(n) .* sum(tmp, 2));
        end
        dC = [dC; dCdM(:)];
    end
    