function [gamma, eta, logZ, emission, edge_list, log_shift] = sumprod_dai(X, G, model)
%SUMPROD_DAI Performs sum-product in a pairwise graphical model using libDAI
%
%   [gamma, eta, logZ, emission, edge_list] = sumprod_dai(X, G, model)
%   [gamma, eta, logZ, emission, edge_list] = sumprod_dai([], G, model) 
%
% Performs the loopy belief propagation on data X with the structure in 
% adjacency matrix G in the CRF specified in model. The final messages are 
% returned in mess. The function also final beliefs over nodes in gamma, the 
% final beliefs over edges in eta, and the log-partition function in logZ.
%
% This function uses libDAI to perform the inference in the label graph. It
% is designed for CRFs with only pairwise potentials.
%
%
% (C) Laurens van der Maaten, 2011
% Delft University of Technology


    % Initialize some variables
    N = size(X, 2);
    K = size(model.A, 1);
    
    % Make sure a node does not message itself
    G = logical(G);
    G(1:N + 1:end) = false;
    no_edges = sum(G(:));
    
    % Compute energy of data-dependent factors
    if ~isempty(X)
        switch model.emission
            case 'gaussian'
                emission = gaussian_emission(X, model);
            case 'multinomial'
                emission = multinomial_emission(X, model);
            otherwise
                error('Unknown emission distribution.');
        end
    else
        emission = zeros(K, size(G, 1));
    end
    log_shift = 0; %max(emission(:));               % log-shift code doesn't work
    exp_emission = exp(emission - log_shift);
    exp_A = exp(model.A - log_shift);
    
    % Construct factor graph in libDAI format
    psi = cell(N + no_edges, 1);
    for n=1:N
        psi{n}.Member = n - 1;
        psi{n}.P = exp_emission(:,n);
    end
    [r, c] = find(G);
    for n=1:no_edges
        psi{n + N}.Member = [r(n) c(n)] - 1;
        psi{n + N}.P = exp_A;
    end
    
    % Perform inference using loopy belief propagation
    [logZ, ~, ~, dai_gamma, dai_eta] = dai(psi, 'BP', '[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=1,inference=SUMPROD,verbose=0]');
    
    % Convert beliefs over nodes into something more manageable
    gamma = zeros(K, N);
    for n=1:length(dai_gamma)
        gamma(:,dai_gamma{n}.Member + 1) = dai_gamma{n}.P;
    end    
    
    % Convert beliefs over edges into something more manageable    
    count = 0;    
    eta = zeros(K, K, no_edges);
    edge_list = zeros(2, no_edges);
    for n=1:length(dai_eta)
        if numel(dai_eta{n}.Member) == 2                % only store beliefs over pairwise edges
            count = count + 1;
            eta(:,:,count) = dai_eta{n}.P;
            edge_list(:,count) = dai_eta{n}.Member';
        end
    end
    assert(count == no_edges);    
    