function model = train_hmm(X, K, emis_type, left_to_right, L)
%TRAIN_HMM Trains a Hidden Markov Model on a data set of time series
%
%   model = train_hmm(X, K, emis_type, left_to_right)
%   model = train_hmm(X, K, 'gaussian_mix', left_to_right, L)
%
% Trains a Hidden Markov Model with K states on the data set of time series
% specified in the cell array X. The Hidden Markov Model uses Gaussian
% emission probabilities. If left_to_right is set to true, the HMM cannot
% return to a previous state (default = false).
% The variable emis_type indicates the type of emission probabilities, and
% can be set to 'gaussian', 'gaussian_mix', or 'multinomial' (default =
% 'gaussian'). L is the number of mixture components for the Gaussian
% mixture emissions (default = 2).
%
%
% (C) Laurens van der Maaten, 2010
% University of California, San Diego


    if ~exist('emis_type', 'var') || isempty(emis_type)
        emis_type = 'gaussian';
    end
    if ~exist('left_to_right', 'var') || isempty(left_to_right)
        left_to_right = false;
    end    
    if ~exist('L', 'var') || isempty(L)
        if strcmpi(emis_type, 'gaussian_mix'), L = 2; else L = 1; end
    end
    if size(X, 1) > size(X, 2)
        X = X';
    end

    % Initialize Hidden Markov Model
    D = size(X{1}, 1);
    model = initialize_hmm(X, emis_type, 'pretrain', K, [], left_to_right, L);
    
    % Initialize some other parameters
    no_series = length(X);
    max_iter = 100;
    iter = 0; L = Inf; old_L = Inf;
    gamma = cell(no_series, 1);
    
    % Perform EM algorithm
    disp(['Training Hidden Markov Model with ' num2str(K) ' states...']);
    while iter < 5 || (L - old_L > 1e-5 && iter < max_iter)
        
        % Create new transition parameters
        new_pi = zeros(size(model.pi));
        new_A  = zeros(size(model.A));
        
        % Loop over all time series
        old_L = L; L = 0; iter = iter + 1;
        for i=1:no_series
            
            % Perform forward-backward algorithm
            Xi = X{i}; N = size(Xi, 2);
            [alpha, beta, rho, emission, gamma{i}] = forward_backward(Xi, model);
            
            % Update prior over states
            new_pi = new_pi + gamma{i}(:,1);
            
            % Update the transition probabilities
            if N > 1
                eps = repmat(realmin, [K K N - 1]);
                log_A = log(max(model.A, realmin));
                for n=2:N
                    eps(:,:,n - 1) = log_A + repmat(log(max(alpha(:,n - 1), realmin)), [1 K]) + repmat(log(max(beta(:,n), realmin))' + emission(:,n)', [K 1]) - rho(n);
                end
                shift = max(eps(:));
                eps = log(sum(exp(eps - shift), 3));
                new_A = new_A + exp(eps + shift);
            end
            
            % Sum log-likelihood
            L = L + sum(rho);
        end
        
        % Update the parameters of the model
        eval(['model = update_' model.emission '(X, model, gamma);']);
        model.pi = new_pi ./ sum(new_pi);
        model.A  = bsxfun(@rdivide, new_A, max(sum(new_A, 2), realmin));
        
        % Print out progress
        disp(['Iteration ' num2str(iter) ': log-likelihood is ' num2str(L)]);
        
        % Perform plotting
        if D == 2 && strcmpi(model.emission, 'gaussian')
            scatter(Xi(1,:), Xi(2,:)); hold on
            for k=1:K
                plot_gauss(model.mu(:,k), model.sigma(:,:,k));
            end
            hold off; drawnow
        end
    end
    